
import signal
import warnings
import itertools
import numpy as np
import typing as T
from sqsgenerator.fallback.attrdict import AttrDict
from operator import attrgetter as attr, itemgetter as item
from sqsgenerator.io import read_settings_file, export_structures
from sqsgenerator.settings import construct_settings, process_settings, defaults
from sqsgenerator.adapters import to_pymatgen_structure, to_ase_atoms, from_pymatgen_structure, from_ase_atoms
from sqsgenerator.core import log_levels, set_core_log_level, pair_sqs_iteration as pair_sqs_iteration_core, \
    SQSResult, symbols_from_z, Structure, make_supercell, IterationMode, pair_analysis, available_species, make_rank, \
    rank_structure, total_permutations

__all__ = [
    'IterationMode',
    'Structure',
    'SQSResult',
    'make_supercell',
    'process_settings',
    'make_result_document',
    'pair_sqs_iteration',
    'sqs_optimize',
    'from_pymatgen_structure',
    'from_ase_atoms',
    'pair_analysis',
    'available_species',
    'read_settings_file',
    'export_structures',
    'make_rank',
    'rank_structure',
    'total_permutations',
    'extract_structures',
    'expand_sqs_results',
    'sqs_analyse'
]

TimingDictionary = T.Dict[int, T.List[float]]
Settings = AttrDict
SQSResultCollection = T.Iterable[SQSResult]


def make_result_document(settings: Settings, sqs_results: T.Iterable[SQSResult],
                         timings: T.Optional[TimingDictionary] = None,
                         fields: T.Tuple[str, ...] = ('configuration',)) -> Settings:
    """
    Converts the ``sqsgenerator.core.SQSResults`` obtained from ``pair_sqs_results`` into a JSON/YAML serializable
    dictionary

    :param settings: the settings dictionary used to compute {sqs_results}
    :type settings: AttrDict
    :param sqs_results: the ``sqsgenerator.core.SQSResults`` calculated by ``pair_sqs_results``
    :type sqs_results: iterable of ``sqsgenerator.core.SQSResults``
    :param timings: a dictionary of thread timing information (default is ``None``)
    :type timings: Dict[int, float]
    :param fields: the fields to include in the document. Can be either *configuration*, *objective* and/or
        *parameters* (default is ``('configuration',)``)
    :type fields: Tuple[str, ...]
    :return: the JSON/YAML serializable document
    :rtype: AttrDict
    """
    allowed_fields = dict(
        configuration=lambda result: symbols_from_z(result.configuration),
        objective=attr('objective'),
        parameters=lambda result: result.parameters(settings.target_objective.shape)
    )

    def make_sqs_result_document(result):
        if len(fields) == 1:
            key = next(iter(fields))
            return allowed_fields[key](result)
        else:
            return {f: allowed_fields[f](result) for f in fields}

    result_document = dict(
        structure=settings.structure,
        configurations={r.rank: make_sqs_result_document(r) for r in sqs_results},
        which=settings.which
    )

    if timings is not None:
        result_document['timings'] = timings
    return Settings(result_document)


def extract_structures(results: Settings, base_structure: T.Optional[Structure] = None) -> T.Dict[int, Structure]:
    """
    Parses a dictionary of results and replaces the generated configuration with the actual structure

    :param results: the dict-like iteration results
    :type results:  AttrDict
    :param base_structure: the structure to which the individual configuration are applied to. If ``None`` it tries
        to extract it from the results document by calling the "structure" attribute (default is ``None``)
    :return: a dictionary with ranks structure objects as values
    :rtype: Dict[int, :py:class:`Structure`]
    """

    structure = results.structure if base_structure is None else base_structure
    which = results.which if base_structure is None else tuple(range(len(structure)))

    raw_data = results['configurations'] if 'configurations' in results else results

    def get_configuration(conf):
        return conf if not isinstance(conf, dict) else conf['configuration']

    structures = {
        rank:
            structure.with_species(get_configuration(conf), which=which)
        for rank, conf in raw_data.items()
    }
    return structures


def pair_sqs_iteration(settings: Settings, minimal: bool = True, similar: bool = False, log_level: str = 'warning') -> \
        T.Tuple[SQSResultCollection, TimingDictionary]:
    """
    Performs an SQS iteration using the {settings} configuration

    :param settings: the dict-like settings used for the iteration. Please refer to the
        `Input parameter <https://sqsgenerator.readthedocs.io/en/latest/input_parameters.html>`_ for further information
    :type settings: AttrDict
    :param minimal: if the result vector contains ``sqsgenerator.core.SQSResult`` objects with different objective
        values, select only those with minimal value of the objective function (default is ``True``)
    :type minimal: bool
    :param similar: in case the result vector contains more than one structure with minimal objective, include also
        degenerate solutions (default is ``False``)
    :type similar: bool
    :param log_level: the log level of the core extension. Valid values are "*trace*", "*debug*", "*info*", "*warning*"
        and "*error*". Please use "*trace*" when you encounter a bug in the core extension and report an issue
        (default is ``"warning"``)
    :type log_level: str
    :return: the minimal configuration and the corresponding Short-range-order parameters as well as timing information
    :rtype: Tuple[Iterable[:py:class:`SQSResult`], Dict[int, float]]
    """
    set_core_log_level(log_levels.get(log_level))

    iteration_settings = construct_settings(settings, False, structure=settings.structure[settings.which])

    interrupted = False

    def handle_sigint(signum, *_):
        assert signum == signal.SIGINT
        nonlocal interrupted
        interrupted = True

    original_handler = signal.getsignal(signal.SIGINT)
    signal.signal(signal.SIGINT, handle_sigint)
    sqs_results, timings = pair_sqs_iteration_core(iteration_settings)
    signal.signal(signal.SIGINT, original_handler)  # restore the old signal handler

    if interrupted:
        warnings.warn('SIGINT received: SQS results may be incomplete')
    best_result = min(sqs_results, key=attr('objective'))
    if minimal:
        sqs_results = list(filter(lambda r: np.isclose(r.objective, best_result.objective), sqs_results))

        if not similar:
            shape = settings.target_objective.shape
            sqs_results = list(
                filter(
                    lambda r: not np.allclose(r.parameters(shape), best_result.parameters(shape)),
                    sqs_results
                )
            )
            sqs_results.append(best_result)

    return sqs_results, timings


def expand_sqs_results(settings: Settings, sqs_results: T.Iterable[SQSResult],
                       timings: T.Optional[TimingDictionary] = None, fields: T.Tuple[str, ...] = ('configuration',),
                       inplace: bool = False) -> Settings:
    """
    Serializes a list of :py:class:`SQSResult` into a JSON/YAML serializable dictionary

    :param settings: the settings used to compute the {sqs_results}
    :type settings: AttrDict
    :param sqs_results: a iterable (list) of :py:class:`SQSResult`
    :type sqs_results: Iterable[:py:class:`SQSResult`]
    :param timings: a dict like information about the performance of the core routines. Keys refer to thread numbers.
        The values represent the average time the thread needed to analyse one configuration in **µs**
        (default is ``None``)
    :type timings: Dict[int, float]
    :param fields: a tuple of fields to include. Allowed fields are "*configuration*", "*objective*", and
        "*parameters*" (default is ``('configuration',)``)
    :type fields: Tuple[str, ...]
    :param inplace: update the the input ``settings`` document instead of creating a new one (default is ``False``)

    """
    dump_include = list(fields)
    if 'configuration' not in dump_include:
        dump_include += ['configuration']
    result_document = make_result_document(settings, sqs_results, fields=tuple(dump_include), timings=timings)

    if inplace:
        settings = settings.copy()
        settings.update(result_document)
        keys_to_remove = {'file_name', 'input_format', 'composition', 'iterations', 'max_output_configurations',
                          'mode', 'threads_per_rank'}
        final_document = {k: v for k, v in settings.items() if k not in keys_to_remove}
    else:
        final_document = result_document

    return Settings(final_document)


def merge(a: dict, b: T.Optional[dict] = None, **kwargs) -> dict:
    if b is None:
        b = {}
    return {**a, **b, **kwargs}


def sqs_optimize(settings: T.Union[Settings, T.Dict], process: bool = True, minimal: bool = True,
                 similar: bool = False, log_level: str = 'warning',
                 fields: T.Tuple[str, ...] = ('configuration', 'parameters', 'objective'),
                 make_structures: bool = False, structure_format: str = 'default') \
        -> T.Tuple[T.Dict[int, T.Dict[str, T.Any]], T.Dict[int, T.Union[float, T.List[float]]]]:
    """
    This function allows to simply generate SQS structures

    Performs a SQS optimization loop. This function is meant for using sqsgenerator through Python. Prefer this function
    over :py:func:`pair_sqs_iteration`. It combines the functionalities of several low-level utility
    function.

        1. Generate default values for {settings} (:py:func:`process_settings`)
        2. Execute the actual SQS optimization loop (:py:func:`pair_sqs_iteration`)
        3. Process, convert the results (:py:func:`make_result_document`)
        4. Build the structures from the optimization results (:py:func:`extract_structures`)

    An example output might look like the following:

        .. code-block:: python

            {
                654984: {
                    'configuration': ['Re', 'W', 'Re', 'W', 'Re', 'W', 'Re', 'W', 'Re', 'W', 'Re', 'W', 'Re', 'W']
                    'objective': 0.0,
                    # only present if make_structures=True
                    # Atoms object if structure_format='ase'
                    'structure': Atoms(symbols='ReWReWReWReWReWReWReW', pbc=True, cell=[6.33, 6.33, 6.33])
                }
            }

    :param settings: the settings used for the SQS optimization
    :type settings: AttrDict or Dict
    :param process: process the input {settings} dictionary (default is ``True``)
    :type process: bool
    :param minimal: Include only configurations with minimum objective function in the results (default is ``True``)
    :type minimal: bool
    :param similar: If the minimum objective is degenerate include also results with same parameters but
        different configuration (default is ``False``)
    :type similar: bool
    :param log_level: set's the log level for the core C++ extension. Possible fields are "*trace*", "*debug*",
        "*info*", "*warning*" and "*error*" (default is ``'warning'``)
    :type log_level: str
    :param fields: output fields included in the result document. Possible fields are "*configuration*", "*parameters*",
        "*objective*" and "*parameters*" (default is ``('configuration',)``)
    :type fields: Tuple[str, ...]
    :param make_structures: build structure objects from the optimization results (default is ``False``
    :type make_structures: bool
    :param structure_format: if {make_structures} was set to ``True`` it specifies the format of the build structures
        (default is ``'default'``)

            - "*default*": :py:class:`Structure`
            - "*pymatgen*" :py:class:`pymatgen.core.Structure`
            - "*ase*": :py:class:`ase.atoms.Atoms`

    :type structure_format: str
    :return: a dictionary with the specified fields as well as timing information. The keys of the result dictionary are
        the permutation ranks of the generated configuration.
    :rtype: Tuple[Dict[int, Dict[str, Any]], Dict[int, Union[float, List[float]]]]

    """
    settings = settings if isinstance(settings, Settings) else AttrDict(settings)
    settings = process_settings(settings) if process else settings

    results, timings = pair_sqs_iteration(settings, minimal=minimal, similar=similar, log_level=log_level)

    result_document = expand_sqs_results(settings, results, timings=timings, fields=fields)
    if make_structures:
        structure_document = extract_structures(result_document)
        converter = dict(default=lambda _: _, ase=to_ase_atoms, pymatgen=to_pymatgen_structure).get(structure_format)
        structure_document = {k: converter(v) for k, v in structure_document.items()}

        result_document = result_document.get('configurations')
        result_document = {rank: merge(result, structure=structure_document[rank]) for rank, result in
                           result_document.items()}
    else:
        result_document = result_document.get('configurations')

    return result_document, timings


def sqs_analyse(settings: T.Union[Settings, T.Dict], structures: T.Iterable[Structure], process: bool = True,
                fields: T.Tuple[str, ...] = ('configuration', 'parameters', 'objective'),
                structure_format: str = 'default', append_structures: bool = False) -> T.Dict[int, T.Dict[str, T.Any]]:
    """
    Uses the given settings {settings} and an iterable of :py:func:`Structure` and compute the short-range-order
    parameters, objective function. By default the {fields} = ('configuration', 'parameters', 'objective') are
    included.

    :param settings: the settings used for the SQS optimization
    :type settings: AttrDict or Dict
    :param structures: an iterable of structures to analyse
    :type structures: Iterable[Union[Structure, :py:class`ase.atoms.Atoms`, :py:class:`pymatgen.core.Structure`]]
    :param process: process the input {settings} dictionary (default is ``True``)
    :type process: bool
    :param fields: output fields included in the result document. Possible fields are "*configuration*", "*parameters*",
        "*objective*" and "*parameters*" (default is ``('configuration', 'parameters', 'objective')``)
    :type fields: Tuple[str, ...]
    :param structure_format: the input format of the items in {structures}
        (default is ``'default'``)

            - "*default*": :py:class:`Structure`
            - "*pymatgen*" :py:class:`pymatgen.core.Structure`
            - "*ase*": :py:class:`ase.atoms.Atoms`
    :type structure_format: str
    :param append_structures: append the initial {structures} to the analysed results (default is ``False``)
    :type append_structures: bool
    :return: a dictionary with the specified fields as well as timing information. The keys of the result dictionary are
        the permutation ranks of the {structures}.
    :rtype: Dict[int, Dict[str, Any]]
    """

    settings = settings if isinstance(settings, Settings) else AttrDict(settings)
    settings = process_settings(settings) if process else settings

    converter = dict(default=lambda _: _, ase=from_ase_atoms, pymatgen=from_pymatgen_structure).get(structure_format)
    slicer = item(settings.which)

    # convert the structures to Structure object and extract the sublattice if needed
    structures = map(lambda st: slicer(converter(st)), structures)
    first_structure = next(iter(structures), None)
    if first_structure is None:
        raise ValueError('The structure input iterable contains no structure')

    analyse_settings = AttrDict(settings.copy())
    analyse_settings.update(structure=first_structure)
    analyse_settings.update(which=defaults.which(analyse_settings))

    # we have consumed the first element of the {structure} iterator we again assemble it
    # we create a copy of the iterator
    structures, structures_copy = itertools.tee(itertools.chain((first_structure,), structures))

    analysed = {
        rank_structure(st): pair_analysis(construct_settings(analyse_settings, False, structure=st))
        for st in structures
    }
    document = expand_sqs_results(analyse_settings, list(analysed.values()), fields=fields).get('configurations')

    if append_structures:
        for structure in structures_copy:
            rank = rank_structure(structure)
            assert rank in document
            document[rank]['structure'] = structure

    return document



"""
Provides utilities for exporting structure objects with ``ase`` and ``pymatgen``.
Tools for reading/writings settings files
"""

import io
import tarfile
import zipfile
import functools
import numpy as np
import typing as T
from frozendict import frozendict
from sqsgenerator.fallback.attrdict import AttrDict
from sqsgenerator.core import Structure, IterationMode
from sqsgenerator.compat import FeatureNotAvailableException
from operator import attrgetter as attr, methodcaller as method
from sqsgenerator.compat import require, have_feature, get_module, Feature
from sqsgenerator.adapters import from_ase_atoms, from_pymatgen_structure, to_pymatgen_structure, to_ase_atoms


F = Feature
known_adapters = (F.ase, F.pymatgen)

compression_to_file_extension = frozendict(zip='zip', bz2='tar.bz2', gz='tar.gz', xz='tar.xz')

output_formats = {
    F.pymatgen: {'cif', 'mcif', 'poscar', 'cssr', 'json', 'xsf', 'prismatic', 'yaml'}
}


def identity(x: T.Any) -> T.Any:
    return x


if have_feature(F.ase):
    from ase.io.formats import all_formats, get_ioformat

    # the underlying ase write function must be capable to cope with file handles
    # therefore we use get_ioformat to get meta-information about the formats, and sort out the unsuitable
    output_formats[F.ase] = {f for f in all_formats.keys() if get_ioformat(f).acceptsfd}


def prepare_handle(fp: T.IO[bytes], feature: Feature, format: str) -> T.Union[T.IO[bytes], T.IO[str]]:
    """
    Sanitizes a file-like by wrapping it in a TextIO if needed

    :param fp: the file-like to sanitize
    :type fp: IO[bytes]
    :param feature: the file writer module "ase" or "pymatgen"
    :type feature: Feature
    :param format: the file format e.g "cif"
    :type format: str
    :return: the file handle
    :rtype: IO[bytes] or IO[str]
    """
    assert format in output_formats[feature]

    if feature == F.pymatgen:
        return io.TextIOWrapper(fp)
    elif feature == F.ase:
        return fp if get_ioformat(format).isbinary else io.TextIOWrapper(fp)


def default_adapter():
    """
    Gets the default writer/reader module to read files

    :return: the default
    :rtype: Feature or None
    """
    for a in known_adapters:
        if have_feature(a): return a
    return None


def supported_formats(feature=None):
    """
    For a given feature ("ase" or "pymatgen") the supported output file formats computed. If None is passed the values
    from ``default_adapter()`` will be used

    :param feature: the writer/reader feature (default is None)
    :type feature: Feature
    :return: the supported output formats
    :rtype: set
    """
    return output_formats.get(feature or default_adapter(), {})


class NonCloseableBytesIO(io.BytesIO):
    """
    Class as workaround. Some ase.io writer functions close the buffer so that we cannot capture their output
    I/O operation on closed buffer Exception. Therefore BytesIO with dummy
    """

    def close(self) -> None:
        pass

    def really_close(self):
        super(NonCloseableBytesIO, self).close()


def capture(f):
    """
    Decorator which captures the written bytes of a "write" function an. It passed the buffer as the first argument
    to the wrapped function

    :param f: a callable returning the written bytes
    :type f: callable
    :return: wrapped function
    """

    @functools.wraps(f)
    def _capture(*args, **kwargs):
        fp = NonCloseableBytesIO()
        f(fp, *args, **kwargs)
        fp.seek(0)
        r = fp.getvalue()
        fp.really_close()
        return r

    return _capture


@require(F.json, F.yaml, F.pickle, condition=any)
def dumps(o: dict, output_format: str = 'yaml') -> bytes:
    """
    Dumps a dict-like object into a byte string, using a backend specified by {output_format}

    :param o: a dict-like object which is serializable by the module specified by {output_format}
    :param output_format: backend used to store (json, pickle, yaml) the object (default is ``"yaml"``
    :type output_format: str
    :return: the dumped content
    :rtype: bytes
    """
    f = F(output_format)
    if not have_feature(f):
        raise FeatureNotAvailableException(f'The package "{format}" is not installed, '
                                           'consider to install it with')

    # for yaml format we create a simple wrapper which captures the output
    def safe_dumps(d, **kwargs):
        buf = io.StringIO()
        get_module(F.yaml).safe_dump(d, buf, **kwargs)
        return buf.getvalue()

    dumpers = {
        F.json: lambda d: get_module(F.json).dumps(d, indent=4),
        F.pickle: lambda d: get_module(F.pickle).dumps(d),
        F.yaml: lambda d: safe_dumps(d, default_flow_style=None)
    }
    content = dumpers[f](o)
    return content if f == F.pickle else content.encode()


@require(F.yaml, F.json, condition=any)
def read_settings_file(path: str, format: str = 'yaml') -> AttrDict:
    """
    Reads a file expecting {format} as the file type. This method does not process the input paramters, but rather
    just reads, them from. To obtain default values for all parameters use :py:func:`process_settings`

    :param path: the file path
    :type path: str
    :param format: the input file-type. Possible formats are *yaml*, *json* and *pickle* (default is ``'yaml'``)
    :type format: str
    :return: the parsed settings
    :rtype: AttrDict
    """
    f = F(format)
    readers = {
        F.json: 'loads',
        F.pickle: 'loads',
        F.yaml: 'safe_load'
    }
    if not have_feature(f):
        raise FeatureNotAvailableException(f'The package "{format}" is not installed, '
                                           'consider to install it with')
    reader = getattr(get_module(f), readers[f])
    try:
        mode = 'r' if f != F.pickle else 'rb'
        with open(path, mode) as settings_file:
            content = settings_file.read()
    except (FileNotFoundError, UnicodeDecodeError):
        raise
    try:
        data = AttrDict(reader(content))
    except Exception as e:
        raise IOError(f'While reading the file "{path}", a "{type(e).__name__}" occurred. '
                      f'Maybe the file has the wrong format. '
                      f'I was expecting a "{format}"-file. '
                      f'You can specify a different input-file format using the "--format" option')
    return data


@require(F.ase)
def read_structure_file_with_ase(fn, **kwargs) -> Structure:
    import ase.io
    return from_ase_atoms(ase.io.read(fn, **kwargs))


@require(F.pymatgen)
def read_structure_file_with_pymatgen(fn, **kwargs) -> Structure:
    import pymatgen.core
    return from_pymatgen_structure(pymatgen.core.Structure.from_file(fn, **kwargs))


@require(F.ase)
def write_structure_file_with_ase(fp, structure: Structure, format, sort=True, **kwargs):
    import ase.io
    ase.io.write(fp, (to_ase_atoms(structure.sorted() if sort else structure),), format=format, **kwargs)


@require(F.pymatgen)
def write_structure_file_with_pymatgen(fp, structure: Structure, format, sort=True, **kwargs):
    fp.write(to_pymatgen_structure(structure.sorted() if sort else structure).to(format, **kwargs))


@require(F.ase, F.pymatgen, condition=any)
def write_structure_file(fp: T.IO[bytes], structure: Structure, format: str, writer: Feature = default_adapter(),
                         **kwargs) -> T.NoReturn:
    """
    Write a ``sqsgenerator.core.Structure`` object into a file, with file format {format} using {writer} as backend

    :param fp: file object
    :type fp: IO[bytes]
    :param structure: the structure to save
    :type structure: Structure
    :param format: file type used. Must be supported by {writer}
    :type format: str
    :param writer: the writer backend (default is ``default_adapter()``)
    :type writer: Feature
    :param kwargs: keyword arguments passed to the backends
    """
    writer_funcs = {
        F.ase: write_structure_file_with_ase,
        F.pymatgen: write_structure_file_with_pymatgen
    }
    fh = prepare_handle(fp, writer, format)
    writer_funcs[writer](fh, structure, format, **kwargs)
    fh.flush()


@require(F.ase, F.pymatgen, condition=any)
def read_structure_from_file(settings: AttrDict) -> Structure:
    """
    Read a structure object from

    :param settings: the settings dictionary
    :type settings: AttrDict
    :return: the Structure object
    :rtype: Structure
    """
    reader = settings.structure.get('reader', 'ase')
    available_readers = set(map(attr('value'), known_adapters))
    if reader not in available_readers:
        raise FeatureNotAvailableException(f'Unknown reader specification "{reader}.  '
                                           f'Available readers are {known_adapters}')
    reader_kwargs = settings.structure.get('args', {})
    reader_funcs = dict(ase=read_structure_file_with_ase, pymatgen=read_structure_file_with_pymatgen)
    return reader_funcs[reader](settings.structure.file, **reader_kwargs)


# Helper to capture the backends output into bytes
dumps_structure = capture(write_structure_file)


def to_dict(settings: dict) -> T.Dict[str, T.Any]:
    """
    Utility method to recursively turn a general dictionary into an JSON/YAML serializable dictionary.
    If a non-trivial object is encountered the function searches for a ``to_dict()`` function. If it has no method
    available to serialize the object a ``TypeError`` is raised. **Attention:** the function serializes ``np.ndarray``
    by calling ``tolist()``. This is not a good idea but fits the needs in this project

    :param settings: a generic dictionary object
    :type settings: dict
    :return: a serializable dict
    :rtype: dict
    """

    identity = lambda _: _  # although bad practice this is readable =)
    converters = {
        int: identity,
        float: identity,
        str: identity,
        bool: identity,
        IterationMode: str,
        np.ndarray: method('tolist')
    }

    def _generic_to_dict(d):
        td = type(d)
        if isinstance(d, (tuple, list, set)):
            return td(map(_generic_to_dict, d))
        elif isinstance(d, dict):
            return dict({k: _generic_to_dict(v) for k, v in d.items()})
        elif td in converters:
            return converters[td](d)
        elif hasattr(d, 'to_dict'):
            return d.to_dict()
        else:
            raise TypeError(f'No converter specified for type "{td}"')

    return _generic_to_dict(settings)


def export_structures(structures: T.Dict[T.Any, T.Any], format: str = 'cif', output_file: str = 'sqs.result',
                      writer: T.Union[Feature,str] = 'ase', compress: T.Optional[str] = None,
                      functor: T.Callable[[T.Any], str] = identity) -> T.NoReturn:
    """
    Writes structures into files. The filename is specified by the keys of {structure} argument. The structures stored
    in the values will be written using the {writer} backend in {format}. If compress is specified the structures will
    be dumped into an archive with name {output_file}. The file-extension is chosen automatically.

    :param structures: a mapping of filenames and :py:class:`Structures`
    :type structures: dict[Any, Structure]
    :param format: output file format (default is ``'cif'``)
    :param output_file: the prefix of the output archive name. File extension is chosen automatically.
        If {compress} is ``None`` this option is ignored (default is ``'sqs.result'``)
    :type output_file: str
    :param writer: the writer backend (default is ``'ase'``)
    :type writer: str
    :param compress: compression algorithm (``'zip'``, ``'gz'``, ``'bz2'`` or ``'xz'``) used to store the structure
        files. If ``None`` the structures are written to plain files (default is ``None``)
    :type compress: str or None
    :param functor: a callable which maps the values of {structures} on a :py:class:`Structure` (default is ``identity = lambda x: x``)
    :type functor: Callable[[Any], Structure]

    """

    writer = Feature(writer) if isinstance(writer, str) else writer

    output_prefix = output_file
    if compress:
        # select the proper file-mode as well as file-name and opening method
        output_archive_file_mode = f'x:{compress}' if compress != 'zip' else 'x'
        output_archive_name = f'{output_prefix}.{compression_to_file_extension.get(compress)}'
        open_ = tarfile.open if compress != 'zip' else zipfile.ZipFile
        archive_handle = open_(output_archive_name, output_archive_file_mode)
    else:
        archive_handle = None

    # helper method, dealing with the compression algorithms
    def write_structure_dump(data: bytes, filename: str):
        if not compress:
            with open(filename, 'wb') as fh:
                fh.write(data)
        else:
            if compress == 'zip':
                assert isinstance(archive_handle, zipfile.ZipFile)
                archive_handle.writestr(filename, data)
            else:
                assert isinstance(archive_handle, tarfile.TarFile)
                with io.BytesIO(data) as buf:
                    tar_info = tarfile.TarInfo(name=filename)
                    tar_info.size = len(data)
                    archive_handle.addfile(tar_info, buf)

    structures = {k: functor(v) for k, v in structures.items()}
    for rank, structure in structures.items():
        filename = f'{rank}.{format}'
        data = dumps_structure(structure, format, writer=writer)  # capture the output from the {writer} backend
        write_structure_dump(data, filename)

    if compress:
        # If we dumped everything into an archive, we have to close the it at the end
        archive_handle.close()



"""
Utilities which are used by the CLI interface commands
"""

import io
import os
import sys
import rich
import click
import functools
import typing as T
from sqsgenerator.io import read_settings_file
from sqsgenerator.commands.help import parameter_help as help
from sqsgenerator.settings import process_settings, parameter_list, BadSettings


def error(message: str, exc_type=click.Abort, raise_exc: bool = True, prefix: T.Optional[str] = None, **kwargs):
    """
    Prints an error message and abort execution by raising an error
    :param message: the error message to print
    :type message: str
    :param exc_type: the exception type to raise (default is `click.Abort`)
    :type exc_type: type
    :param raise_exc: raise an exception (default is `True`)
    :type raise_exc: True
    :param prefix: a bold red prefix to the error message
    :type prefix: str
    :param kwargs: keyword args are forwarded `to click.echo`
    """
    message = pretty_print(message, show=False)
    if prefix is not None:
        prefix = click.style(prefix, fg='red', bold=True, underline=True)
        message = f'{prefix}: {message}'
    click.echo(message, **kwargs)
    if raise_exc:
        raise exc_type(message)


def pretty_print(*objects: T.Any, show: bool = True, paginate: str = 'auto', **kwargs):
    """
    Formats an object nicely. Tries to use "rich" if available. Falls back to pprint.pprint alternatively
    :param objects: the objects to print
    :param show: if `show` is true the output will be printed to `sys.stdout` (default is `True`)
    :type show: bool
    :param paginate: paginated output, allowed is `True`, `False` and `'auto'` (default is `'auto'`)
    :type paginate: bool or str
    :param kwargs: keyword args passed on to rich.console.print  or pprint.pprint
    :return: the formatted output if `show` is `False` else `None` is returned
    :rtype: str of None
    """
    buf = io.StringIO()
    console = rich.get_console()
    with console.capture() as capture:
        console.print(*objects, **kwargs)
    buf.write(capture.get())
    console_height = console.height

    string = buf.getvalue()
    if paginate == 'auto':
        paginate = string.count(os.linesep) >= console_height
    printer = functools.partial(click.echo, file=sys.stdout, nl=False) if not paginate else click.echo_via_pager
    return printer(buf.getvalue()) if show else buf.getvalue()


def make_help_link(parameter: str) -> str:
    """
    Formats a help link pointing to the input parameter documentation
    :param parameter: the input parameter name
    :type parameter: str
    :return: the helplink
    :rtype: str
    """
    section_permalink = parameter.replace('_', '-')
    base_url = 'https://sqsgenerator.readthedocs.io/en/latest'
    help_page = 'input_parameters.html'
    return f'{base_url}/{help_page}#{section_permalink}'


def exit_on_input_parameter_error(f):
    """
    Decorator: Wraps a @parameter decorated function -> catches an eventual `BadSettings` error -> creates a help-link
    for the parameter -> redirects the `BadSettings` into `click.Abort` to stop CLI execution
    """

    @functools.wraps(f)
    def _inner(*args, **kwargs):
        try:
            result = f(*args, **kwargs)
        except BadSettings as exc:
            prefix = type(exc).__name__
            raise_first = exc.parameter is None
            error(exc.message, prefix=prefix, raise_exc=raise_first, nl=False)
            error(f'Maybe the documentation can help you: {make_help_link(exc.parameter)}',
                  prefix=prefix, raise_exc=not raise_first, nl=False)
        else:
            return result

    return _inner


def click_settings_file(process=None, default_name='sqs.yaml', ignore=()):
    """
    Decorator-Factory: A lot of the commands need an input YAML file -> creates a decorator which parses a settings
    file and takes an input format as parameter. The decorator processes the specified file
    :param process: process (parsed) the settings dictionary with sqsgenerator.core (default is `None`)
    :type process: Iterable[str] or None
    :param default_name: if no file path is specified the function will search {default_name} (default is `"sqs.yaml"`)
    :type default_name: str
    :param ignore: list of parameters names which are ignored
    :return: the function wrapped by a `click.option` {input_format} and a `click.argument` {filename} decorator
    :rtype: callable
    """

    def _decorator(f: T.Callable):

        @functools.wraps(f)
        @click.argument('filename', type=click.Path(exists=True), default=default_name)
        @click.option('--input-format', '-if', type=click.Choice(['yaml', 'json', 'pickle']), default='yaml',
                      help=help.input_format)
        def _dummy(*args, filename=default_name, input_format='yaml', **kwargs):
            settings = read_settings_file(filename, format=input_format)
            settings['file_name'] = filename
            settings['input_format'] = input_format
            nonlocal process
            if process is not None:
                param_list = parameter_list()
                if 'all' in process:
                    process = param_list

                settings = process_settings(settings, params=process, ignore=ignore)

            return f(settings, *args, **kwargs)

        return exit_on_input_parameter_error(_dummy)

    return _decorator


pretty_format = functools.partial(pretty_print, show=False)



import numbers
import itertools
import numpy as np
import typing as T
from operator import attrgetter as attr
from sqsgenerator.core.core import Structure as Structure_


class Structure(Structure_):
    """
    Structure class used to store structural information. This class is used by the core extension and is a wrapper
    around the extension internal ``sqsgenerator.core.core.Structure`` class. The class is designed to be array like,
    therefore it does not provide any setter functions. Internally this class uses **fractional coordinates** to
    represent structural information.
    """

    @classmethod
    def from_extension_class(cls, o: Structure_):
        return cls(o.lattice.copy(), o.frac_coords.copy(), list(map(attr('symbol'), o.species)), o.pbc)

    def __init__(self, lattice: np.ndarray, frac_coords: np.ndarray, symbols: T.List[str],
                 pbc: T.Tuple[bool, bool, bool] = (True, True, True)):
        """
        Constructs a new structure from raw data

        :param lattice: the (3x3) lattice matrix. The three **rows** are interpreted as the lattice vectors **a**,
            **b** and **c**.
        :type lattice: np.ndarray
        :param frac_coords: the **fractional coordinates** of the lattice positions as (3xN) array. Each **row**
            will be treated as a lattice position
        :type frac_coords: np.ndarray
        :param symbols: a list of strings of length N specifying the atomic species which occupy the lattice positions
        :type symbols: List[str]
        :param pbc: the coordinate axes for which **periodic boundary conditions** should be applied.
            **Do not pass a value here**. This feature is not yet implemented (default is ``(True, True, True)``)
        :type pbc: Tuple[bool, bool, bool]
        :raises ValueError: if length of {frac_coords} and length of {symbols} do not match

        """
        super(Structure, self).__init__(lattice, frac_coords, symbols, pbc)
        self._symbols = np.array(list(map(attr('symbol'), self.species)), dtype='<U3')  # Uut is the longest symbol
        self._numbers = np.fromiter(map(attr('Z'), self.species), dtype=int, count=self.num_atoms)
        self._unique_species = set(np.unique(self._symbols))

    @property
    def symbols(self) -> np.ndarray:
        """
        A ``numpy.ndarray`` storing the symbols of the atomic species. E.g "*Fe*", "*Cr*", "*Ni*"

        :return: the array of symbols
        :rtype: numpy.ndarray
        """
        return self._symbols

    @property
    def numbers(self) -> np.ndarray:
        """
        The ordinal numbers of the atoms sitting on the lattice positions

        :return: array of ordinal numbers
        :rtype: numpy.ndarray
        """
        return self._numbers

    @property
    def num_unique_species(self) -> int:
        """
        The number of unique elements in the structure object

        :return: the number of elements
        :rtype: int
        """
        return len(self._unique_species)

    @property
    def unique_species(self) -> T.Set[str]:
        """
        A set of symbols of the occurring species

        :return: a set containing the symbol of the occurring species
        :rtype: Set[str]
        """
        return self._unique_species

    def without_vacancies(self):
        """
        Removes vacancies from the structure. Removes all lattice sites which are occupied by the "0" species

        :return: a structure with no vacancies
        :rtype: Structure
        """
        return self[~(self.numbers == 0)]

    def __len__(self):
        return self.num_atoms

    def __repr__(self):
        def group_symbols():
            for species, same in itertools.groupby(self._symbols):
                num_same = len(list(same))
                yield species if num_same == 1 else f'{species}{num_same}'

        formula = ''.join(group_symbols())
        return f'Structure({formula}, len={self.num_atoms})'

    def __eq__(self, other):
        if not isinstance(other, Structure):
            return False
        same_species = all(this_num == other_num for this_num, other_num in zip(self.numbers.flat, other.numbers.flat))
        same_lattice = np.allclose(self.lattice, other.lattice)
        same_coords = np.allclose(self.frac_coords, other.frac_coords)
        return all((same_species, same_lattice, same_coords))

    def sorted(self):
        """
        Creates a new structure, where the lattice positions are ordered by the ordinal numbers of the occupying species

        :return: the sorted Structure
        :rtype: Structure
        """
        return self.from_extension_class(super(Structure, self).sorted())

    def rearranged(self, order):
        return self.from_extension_class(super(Structure, self).rearranged(order))

    def __getitem__(self, item):
        if isinstance(item, numbers.Integral):
            if item < -self.num_atoms or item >= self.num_atoms:
                raise IndexError('Index out of range.')
            return self.species[item]

        if isinstance(item, (list, tuple, np.ndarray)):
            indices = np.array(item)
            if indices.dtype == bool:
                indices = np.argwhere(indices).flatten()
            elif indices.dtype != int:
                raise TypeError('Only integer numbers cann be used for slicing')
                # boolean mask slice
        elif isinstance(item, slice):
            indices = np.arange(self.num_atoms)[item]
        else:
            raise TypeError(f'Structure indices must not be of type {type(item)}')
        return Structure(self.lattice, self.frac_coords[indices], self.symbols[indices])

    def to_dict(self) -> dict:
        """
        Serializes the object into JSON/YAML serializable dictionary

        :return: the JSON/YAML serializable dictionary
        :rtype: Dict[str, Any]
        """
        return structure_to_dict(self)

    def slice_with_species(self, species: T.Iterable[str], which: T.Optional[T.Iterable[int]] = None):
        """
        Creates a new structure containing the lattice positions specified by {which}. The new structure
        is occupied by the atomic elements specified in {species}. In case {which} is ``None`` all lattice positions
        are assumed to be occupied with a new {species}. The new structure will only contain lattice positions specified
        in {which}.

        :param species: the atomic species specified by their symbols
        :type species: Iterable[str]
        :param which: the indices of the lattice positions to choose (default is ``None``)
        :type which: Optional[Iterable[int]]
        :return: the (subset) structure with new species
        :rtype: Structure
        :raises ValueError: if length of which is < 1 or length of {which} and {species} does not match
        """
        which = which or tuple(range(self.num_atoms))
        return self.with_species(species, which=which)[which]

    def with_species(self, species, which=None):
        """
        Creates a new structure containing the lattice positions specified by {which}. The new structure
        is occupied by the atomic elements specified in {species}. In case {which} is ``None`` all lattice positions
        are assumed to be occupied with a new {species}. The new structure will only all lattice positions of the
        current structure, while on the positions specified by {which} are occupied with {species}

        :param species: the atomic species specified by their symbols
        :type species: Iterable[str]
        :param which: the indices of the lattice positions to choose (default is ``None``)
        :type which: Optional[Iterable[int]]
        :return: the structure with new species
        :rtype: Structure
        :raises ValueError: if length of which is < 1 or length of {which} and {species} does not match
        """

        which = which or tuple(range(len(self)))
        species = list(species)
        if len(species) < 1:
            raise ValueError('Cannot create an empty structure')
        if len(which) != len(species):
            raise ValueError('Number of species does not match the number of specified atoms')
        new_symbols = self.symbols.copy()
        new_symbols[np.array(which)] = species
        return Structure(self.lattice, self.frac_coords, new_symbols.tolist())


def structure_to_dict(structure: Structure):
    structure = structure.without_vacancies()
    return dict(
        lattice=structure.lattice.tolist(),
        coords=structure.frac_coords.tolist(),
        species=structure.symbols.tolist()
    )


def make_supercell(structure: Structure, sa: int = 1, sb: int = 1, sc: int = 1) -> Structure:
    """
    Creates a supercell of structure, which is repeated {sa}, {sb} and {sc} times

    :param structure: the structure to replicate
    :type structure: Structure
    :param sa: number of repetitions in directions of the first lattice vector (default is ``1``)
    :type sa: int
    :param sb: number of repetitions in directions of the second lattice vector (default is ``1``)
    :type sb: int
    :param sc: number of repetitions in directions of the third lattice vector (default is ``1``)
    :type sc: int
    :return: the supercell structure
    :rtype: Structure
    """
    sizes = (sa, sb, sc)
    num_cells = np.prod(sizes)
    num_atoms_supercell = structure.num_atoms * num_cells
    scale = np.diag(sizes).astype(float)
    iscale = np.linalg.inv(scale)
    supercell_lattice = scale @ structure.lattice

    scaled_fc = structure.frac_coords @ iscale.T

    def make_translation_vector(a, b, c):
        return np.tile(np.array([a, b, c]) @ iscale, [structure.num_atoms, 1])

    supercell_coords = np.vstack([
        scaled_fc + make_translation_vector(*shift)
        for shift in itertools.product(*map(range, sizes))
    ])

    supercell_species = np.tile(structure.symbols, num_cells).tolist()

    assert supercell_coords.shape == (num_atoms_supercell, 3)
    assert len(supercell_species) == num_atoms_supercell
    structure_supercell = Structure(supercell_lattice, supercell_coords, supercell_species, (True, True, True))
    return structure_supercell



"""
Compute quantities which help optimizing the input settings
"""

import time
import click
import numpy as np
import collections
from math import isclose
from sqsgenerator.commands.help import command_help as c_help
from sqsgenerator.settings import construct_settings, build_structure
from sqsgenerator.commands.common import click_settings_file, pretty_print
from sqsgenerator.core import total_permutations as total_permutations_, default_shell_distances, \
    rank_structure as rank_structure_core, IterationMode, pair_sqs_iteration


@click.command('total-permutations', help=c_help.compute.total_permutations)
@click_settings_file({'structure', 'mode', 'iterations', 'composition'})
def total_permutations(settings):
    structure = build_structure(settings.composition, settings.structure[settings.which])
    permutations = total_permutations_(structure) \
        if settings.mode == IterationMode.systematic else settings.iterations
    pretty_print(permutations)
    return permutations


@click.command('shell-distances', help=c_help.compute.shell_distances)
@click_settings_file({'atol', 'rtol', 'structure', 'which', 'composition'})
def shell_distances(settings):
    structure = build_structure(settings.composition, settings.structure[settings.which])
    distances = default_shell_distances(structure, settings.atol, settings.rtol)
    pretty_print(distances)
    return distances


@click.command('rank', help=c_help.compute.rank)
@click_settings_file({'structure'})
def rank_structure(settings):
    rank = pretty_print(rank_structure_core(settings.structure))
    pretty_print(rank)
    return rank


@click.group('compute', help=c_help.compute.command)
def compute():
    pass


def format_seconds(seconds: float) -> str:
    """
    Runtimes can be really long, due to exponentially growing configurational space, therefore those h
    """
    minute = 60
    hour = 60*minute
    day = 24*hour
    year = 365 * day
    units = collections.OrderedDict([
        (230000000*year, 'galactic year'),
        (1000*year, 'millenium'),
        (10*year, 'decade'),
        (year, 'year'),
        (7*day, 'week'),
        (day, 'day'),
        (hour, 'hour'),
        (minute, 'minute')
    ])
    result = []
    for uval, label in units.items():
        if seconds > uval:
            q, r = divmod(seconds, uval)
            result.append(f'{q:.0f} {label}{"s" if not isclose(q, 1.0) else ""}')
            seconds -= q*uval
    if not isclose(seconds, 0.0):
        result += ['and', f'{seconds:.3f} seconds']
    return ' '.join(result)


@click.command('estimated-time', help=c_help.compute.estimated_time)
@click.option('--verbose', '-v', is_flag=True, default=False, help='Print very detailed infos')
@click_settings_file('all')
def estimate_time(settings, verbose):
    have_random_mode = settings.mode == IterationMode.random
    structure = build_structure(settings.composition, settings.structure[settings.which])
    num_iterations = settings.iterations if have_random_mode else total_permutations_(structure)
    num_test_iterations = 100000
    default_guess_settings = dict(iterations=num_test_iterations, mode=IterationMode.random)
    settings.update(default_guess_settings)
    iteration_settings = construct_settings(settings, False, structure=structure)

    t0 = time.time()
    _, timings = pair_sqs_iteration(iteration_settings)
    tend = time.time()
    duration = tend - t0
    num_rank = 0
    thread_timings = timings[num_rank]
    average_thread_time = np.average(thread_timings)
    num_threads = len(thread_timings)
    pure_loop_time = 1e-6*average_thread_time*num_test_iterations/num_threads
    overhead = duration - pure_loop_time
    estimated_seconds = num_iterations * average_thread_time / (num_threads * 1e6)

    pretty_print(f'It will take me roughly {format_seconds(estimated_seconds)} '
                 f'to compute {num_iterations} iterations (on {num_threads} threads)')
    if verbose:
        pretty_print(f'The test was carried out with {num_threads} threads')
        pretty_print(f'On average it takes me {average_thread_time / num_threads:.3f} µs '
                     f'to analyze one configuration ({average_thread_time:.3f} µs per thread)')
        pretty_print(f'This is a throughput of {1e6 / (average_thread_time / num_threads):.0f} '
                     f'configurations/s ({1e6 / average_thread_time:.0f} configurations/s per thread)')
        pretty_print(f'The initialization/finalization overhead was {overhead:.2f} seconds')

    return estimated_seconds


compute.add_command(rank_structure, 'rank')
compute.add_command(shell_distances, 'shell-distances')
compute.add_command(total_permutations, 'total-permutations')
compute.add_command(estimate_time, 'estimated-time')



"""
A data file storing the CLI's help strings
"""

from sqsgenerator.io import default_adapter
from sqsgenerator.fallback.attrdict import AttrDict


command_help = AttrDict(dict(
    analyse='Calculate SRO parameters of existing structures. Searches by default for a file named "sqs.result.yaml".',
    compute=dict(
        command='Compute quantities to determine how computationally demanding you input file is',
        rank='Calculate a unique number identifying the structure in the input file',
        shell_distances='Calculate the default radii of the coordination shells',
        total_permutations='The number of iterations carried out. Useful for for "systematic" mode',
        estimated_time='Estimate the runtime of your current settings'
    ),
    export='Export sqsgenerator\'s results into structure files',
    params=dict(
        command='Utilities to check the parameters that are going to be used in the SQS iteration',
        show='Display all or certain input parameters',
        check='Check if your input settings will work'
    ),
    iteration='Perform a SQS iteration'
))

parameter_help = AttrDict(dict(
    filename='A path to a file readable by sqsgenerator',  # TODO: Add help links here
    input_format='The filetype of {filename}',
    dump_params='Dump input parameters - Append them to the newly generated output',
    output_format='Output format type',
    format='Output format for the structure file(s). The format must be supported by the specified backend',
    writer=f'Backend for writing the structure files. Default is "{default_adapter()}"',
    compress='Gather the output structure files in an compressed archive',
    output_file='A file name prefix. The file extension is chosen automatically',
    param='Display only this input parameter(s)',
    inplace='Dump the computed input parameters in the input settings file',
    dump_format='Format used to store sqsgenerator\'s results',
    dump_include='Include field in sqsgenerator\'s results',
    dump='Dump sqsgenerator\'s results in a file',
    log_level='log level of the core extension. Should be "trace" used for error reporting',
    export='Export the results obtained from the SQS iteration as structure files',
    minimal='Include only configurations with minimum objective function in the results',
    similar='If the minimum objective is degenerate include also results with same parameters but different configuration'
))




"""
A right-favoring Mapping merge.
"""
try:
    from collections.abc import Mapping
except ImportError:
    from collections import Mapping


__all__ = ['merge']


def merge(left, right):
    """
    Merge two mappings objects together, combining overlapping Mappings,
    and favoring right-values

    left: The left Mapping object.
    right: The right (favored) Mapping object.

    NOTE: This is not commutative (merge(a,b) != merge(b,a)).
    """
    merged = {}

    left_keys = frozenset(left)
    right_keys = frozenset(right)

    # Items only in the left Mapping
    for key in left_keys - right_keys:
        merged[key] = left[key]

    # Items only in the right Mapping
    for key in right_keys - left_keys:
        merged[key] = right[key]

    # in both
    for key in left_keys & right_keys:
        left_value = left[key]
        right_value = right[key]

        if (isinstance(left_value, Mapping) and
                isinstance(right_value, Mapping)):  # recursive merge
            merged[key] = merge(left_value, right_value)
        else:  # overwrite with right value
            merged[key] = right_value

    return merged


"""
Provides utility functions and accsess to optional modules
"""

import sys
import logging
import operator
import functools
import frozendict
import typing as T
from enum import Enum


class Feature(Enum):
    ase = 'ase'
    pymatgen = 'pymatgen'
    rich = 'rich'
    pyiron = 'pyiron'
    json = 'json'
    yaml = 'yaml'
    pickle = 'pickle'


class FeatureNotAvailableException(Exception):
    pass


__features = None


def get_module(feature: Feature):
    return sys.modules[feature.value]


def is_initialized():
    return __features is not None


def _is_feature_available(feature: Feature):
    try:
        __import__(feature.value)
    except ImportError:
        return False
    else:
        return True


def _check_features():
    global __features
    if __features is None:
        self_module = __import__(__name__)
        __features = frozendict.frozendict(map(lambda f: (f.value, _is_feature_available(f)), Feature))
        for feat in Feature:
            getter = functools.partial(operator.itemgetter(feat.value), __features)
            setattr(self_module, f'have_{feat.value}', getter)
            message = f'Feature "{feat.value}" was ' + ('found' if __features[feat.value] else 'not found')
            logging.getLogger('compat.check_features').info(message)


def have_feature(feature: T.Union[Feature, str]) -> bool:
    assert is_initialized()
    feature_value = feature.value if isinstance(feature, Feature) else feature
    return feature_value in __features and __features[feature_value]


def require(*features: Feature, condition=all):

    def _decorator(f: T.Callable):
        assert is_initialized()

        @functools.wraps(f)
        def _wrapper(*args, **kwargs):
            if not condition(map(have_feature, features)):
                raise FeatureNotAvailableException(features)

            return f(*args, **kwargs)

        return _wrapper

    return _decorator


def have_mpi_support():
    import sqsgenerator.core.core
    return 'mpi' in sqsgenerator.core.core.__features__


def available_features():
    return tuple(feature.value for feature in Feature if have_feature(feature))


def available_features_with_version():

    def default_version(*_):
        return ''

    def module_version_attr(f):
        feature = Feature(f)
        mod = get_module(feature)
        return f'-{mod.__version__}' if hasattr(mod, '__version__') else default_version(feature)

    version_getters = {
        'ase': module_version_attr,
        'pymatgen': module_version_attr,
        'pyiron': module_version_attr,
        'yaml': module_version_attr,
        'mpi4py': module_version_attr
    }
    return tuple(f'{feature_str}{version_getters.get(feature_str, default_version)(feature_str)}'
                 for feature_str in available_features())


_check_features()



import collections
import numpy as np
from operator import itemgetter as item
from sqsgenerator.fallback.attrdict import AttrDict
from sqsgenerator.settings.utils import build_structure, to_internal_composition_specs
from sqsgenerator.settings.functional import const, if_
from sqsgenerator.core import IterationMode, Structure, default_shell_distances as default_shell_distances_core, \
    compute_prefactors as default_compute_prefactors


ATOL = 1e-3
RTOL = 1e-5


def num_shells(settings: AttrDict):
    return len(settings.shell_weights)


def num_species(settings: AttrDict):
    return build_structure(settings.composition, settings.structure[settings.which]).num_unique_species


def random_mode(settings) -> bool:
    return settings.mode == IterationMode.random


def default_which(settings: AttrDict):
    structure: Structure = settings.structure
    return tuple(range(structure.num_atoms))


def default_composition(settings: AttrDict):
    structure: Structure = settings.structure
    which = settings.which
    return dict(collections.Counter(structure[which].symbols.tolist()))


def default_shell_distances(settings: AttrDict):
    structure = settings.structure[settings.which].sorted()
    return default_shell_distances_core(structure, settings.atol, settings.rtol)


def default_shell_weights(settings: AttrDict):
    return {i: 1.0/i for i, distance in enumerate(settings.shell_distances[1:], start=1)}


def default_pair_weights(settings: AttrDict):
    structure = build_structure(settings.composition, settings.structure[settings.which])
    per_shell_weights = (~np.eye(structure.num_unique_species, dtype=bool)).astype(float) * 0.5
    stack = [weight*per_shell_weights for _, weight in sorted(settings.shell_weights.items(), key=item(0))]
    return np.stack(stack)


def default_target_objective(settings: AttrDict):
    structure = build_structure(settings.composition, settings.structure[settings.which])
    return np.zeros((num_shells(settings), structure.num_unique_species, structure.num_unique_species))


def default_prefactors(settings: AttrDict):
    internal_composition_spec = to_internal_composition_specs(settings.composition, settings.structure)
    return default_compute_prefactors(settings.structure[settings.which], internal_composition_spec,
                                      dict(settings.shell_weights), settings.atol, settings.rtol)


defaults = AttrDict(
    dict(
        atol=const(ATOL),
        rtol=const(RTOL),
        mode=const(IterationMode.random),
        iterations=if_(random_mode)(100000)(-1),
        max_output_configurations=const(10),
        which=default_which,
        composition=default_composition,
        shell_distances=default_shell_distances,
        shell_weights=default_shell_weights,
        pair_weights=default_pair_weights,
        target_objective=default_target_objective,
        prefactors=default_prefactors,
        prefactor_mode=const('set'),
        threads_per_rank=const([-1])
    )
)



"""
Utility functions to convert ``sqsgenerator.core.Structure`` to ``ase.Atoms`` and ``pymatgen.core.Structure``
and vice versa
"""


import numpy as np
import typing as T
from operator import attrgetter as attr
from sqsgenerator.core import Structure
from sqsgenerator.compat import Feature as F, require


def transpose(iterable: T.Iterable):
    return zip(*iterable)


@require(F.pymatgen)
def to_pymatgen_structure(structure: Structure):
    """
    Convert structure {structure} to a ``pymatgen.core.Structure`` object

    :param structure: the structure to convert
    :type structure: Structure
    :return: the ``pymatgen.core.Structure`` object
    :rtype: pymatgen.core.Structure
    """
    from pymatgen.core import Structure as PymatgenStructure, Lattice as PymatgenLattice
    structure = structure.without_vacancies()
    symbols = list(map(attr('symbol'), structure.species))
    lattice = PymatgenLattice(structure.lattice)
    return PymatgenStructure(lattice, symbols, structure.frac_coords, coords_are_cartesian=False, validate_proximity=True)


@require(F.ase)
def to_ase_atoms(structure: Structure):
    """
    Convert structure {structure} to a ``ase.Atoms`` object

    :param structure: the structure to convert
    :type structure: Structure
    :return: the ``ase.Atoms`` object
    :rtype: ase.Atoms
    """
    from ase import Atoms
    structure = structure.without_vacancies()
    numbers = list(map(attr('Z'), structure.species))
    return Atoms(cell=structure.lattice, scaled_positions=structure.frac_coords, numbers=numbers, pbc=structure.pbc)


@require(F.pymatgen)
def from_pymatgen_structure(structure) -> Structure:
    """
    Convert the ``pymatgen.core.Structure`` {structure} object to a ``sqsgenerator.core.Structure`` object

    :param structure: the structure to convert
    :type structure: pymatgen.core.Structure
    :return: the structure object
    :rtype: Structure
    """
    lattice = structure.lattice.matrix
    data = map(lambda site: (site.species_string, site.frac_coords), structure)
    species, frac_coords = transpose(data)
    return Structure(np.array(lattice), np.array(frac_coords), species, (True, True, True))


@require(F.ase)
def from_ase_atoms(atoms) -> Structure:
    """
    Convert the ``ase.Atoms`` {structure} object to a ``sqsgenerator.core.Structure`` object

    :param atoms: the structure to convert
    :type atoms: ase.Atoms
    :return: the structure object
    :rtype: Structure
    """
    lattice = np.array(atoms.cell)
    frac_coords = np.array(atoms.get_scaled_positions()).copy()
    if tuple(atoms.pbc) != (True, True, True):
        raise RuntimeWarning("At present I can only handle fully periodic structures. "
                             "I'will overwrite ase.Atoms.pbc setting to (True, True, True)")
    return Structure(lattice, frac_coords, list(atoms.symbols), (True, True, True))




"""
Utility command to export the internal dict-like results into structure files
"""

import os
import click
from sqsgenerator.public import extract_structures
from sqsgenerator.fallback.attrdict import AttrDict
from sqsgenerator.settings.readers import read_structure
from sqsgenerator.compat import Feature as F, have_feature
from sqsgenerator.commands.common import click_settings_file, error
from sqsgenerator.commands.help import command_help as c_help, parameter_help as help
from sqsgenerator.io import export_structures as export_structures_io, supported_formats, compression_to_file_extension


def export_structures(settings, result_document, output_file=None, format='cif', writer='ase', compress=None):
    """
    Helper function for `sqsgenerator.io.export_structures`. Checks if the requested backend is available and
    the output format is supported by the backend
    """

    # generate a output file name
    #  - in case it was called from the CLI @click_settings_file injected a 'file_name' key into settings
    #  - we use this key if {output_file} was not explicitly specified
    #  - if neither is the case we fall back to "sqs"
    output_prefix = output_file \
        if output_file is not None \
        else (os.path.splitext(settings.file_name)[0] if 'file_name' in settings else 'sqs')

    writer = F(writer)
    # check backend is available
    if not have_feature(writer):
        error(f'I cannot use "{writer.value}". It seems it is not installed', prefix='FeatureError')

    # check if backend supports {format}
    if format not in supported_formats(writer):
        error(f'{writer.value} does not support the format "{format}". '
              f'Supported formats are {supported_formats(writer)}', prefix='FeatureError')

    structures = extract_structures(result_document)
    export_structures_io(structures, format=format, output_file=output_prefix, writer=writer, compress=compress)


@click.command('export', help=c_help.export)
@click.option('--format', '-f', type=click.STRING, default='cif', help=help.format)
@click.option('--writer', '-w', type=click.Choice(['ase', 'pymatgen']), default='ase', help=help.writer)
@click.option('--compress', '-c', type=click.Choice(list(compression_to_file_extension)), help=help.compress)
@click.option('--output', '-o', 'output_file', type=click.Path(dir_okay=False, allow_dash=True), help=help.output_file)
@click_settings_file(process=None, default_name='sqs.result.yaml')
def export(settings, format='cif', writer='ase', compress=None, output_file='sqs.result'):

    result_document = AttrDict(settings)
    needed_keys = {'structure', 'configurations'}

    # check if data needed for export is there at all
    if not all(map(lambda k: k in result_document, needed_keys)):
        error(f'Result document must at least contain the following keys: {needed_keys}', prefix='KeyError')

    result_document['structure'] = read_structure(result_document)
    export_structures(settings, result_document, format=format, writer=writer, compress=compress,
                      output_file=output_file)



"""
Module containing all the routines to process input settings, from the dict-like configuration. Those routines are
intended for internal use only, and are implicitly exported by the ``process_settings`` function.
Each of the function with a @parameter decorator are registered in the private module level {__parameter_registry} 
dict, which makes the visible to the ``process_settings`` function
"""

import numbers
import numpy as np
import typing as T
import collections
import collections.abc
from functools import partial
from operator import itemgetter as item
from sqsgenerator.fallback.attrdict import AttrDict
from sqsgenerator.io import read_structure_from_file
from sqsgenerator.settings.exceptions import BadSettings
from sqsgenerator.core import IterationMode, Structure, make_supercell
from sqsgenerator.compat import Feature, have_mpi_support, have_feature
from sqsgenerator.adapters import from_ase_atoms, from_pymatgen_structure
from sqsgenerator.settings.functional import parameter as parameter_, if_, isa
from sqsgenerator.settings.defaults import defaults, random_mode, num_shells, num_species
from sqsgenerator.settings.utils import ensure_array_shape, ensure_array_symmetric, convert, int_safe, \
    build_structure, to_internal_composition_specs

__parameter_registry = collections.OrderedDict({})

# the parameter decorator registers all the "processor" function with their names in __parameter_registry
# the ordering will be according to their definition in this file
parameter = partial(parameter_, registry=__parameter_registry)


@parameter('atol', default=defaults.atol)
def read_atol(settings: AttrDict):
    if not isinstance(settings.atol, float) or settings.atol < 0:
        raise BadSettings('The absolute tolerance can be only a positive floating point number')
    return settings.atol


@parameter('rtol', default=defaults.rtol)
def read_rtol(settings: AttrDict):
    if not isinstance(settings.rtol, float) or settings.rtol < 0:
        raise BadSettings('The relative tolerance can be only a positive floating point number')
    return settings.rtol


@parameter('mode', default=defaults.mode, required=True)
def read_mode(settings: AttrDict):
    if isinstance(settings.mode, IterationMode):
        return settings.mode
    if settings.mode not in IterationMode.names:
        raise BadSettings(f'Unknown iteration mode "{settings.mode}". '
                          f'Available iteration modes are {list(IterationMode.names.keys())}')
    return IterationMode.names[settings.mode]


@parameter('iterations', default=defaults.iterations, required=if_(random_mode)(True)(False))
def read_iterations(settings: AttrDict):
    num_iterations = convert(settings.iterations, converter=int_safe,
                             message=f'Cannot convert "{settings.iterations}" to int')
    if num_iterations < 0:
        raise BadSettings('"iterations" must be positive')
    return num_iterations


@parameter('max_output_configurations', default=defaults.max_output_configurations)
def read_max_output_configurations(settings: AttrDict):
    num_confs = convert(settings.max_output_configurations, converter=int_safe,
                        message=f'Cannot convert "{settings.max_output_configurations}" to int')
    if num_confs < 0:
        raise BadSettings('"max_output_configurations" must be positive')
    return num_confs


@parameter('structure')
def read_structure(settings: AttrDict) -> Structure:
    needed_fields = {'lattice', 'coords', 'species'}
    s = settings.structure
    structure = None
    if isinstance(s, Structure):
        structure = s

    if have_feature(Feature.ase) and structure is None:
        from ase import Atoms
        if isinstance(s, Atoms):
            structure = from_ase_atoms(s)
    if have_feature(Feature.pymatgen) and structure is None:
        from pymatgen.core import Structure as PymatgenStructure, PeriodicSite
        if isinstance(s, PymatgenStructure):
            structure = from_pymatgen_structure(s)
        elif isinstance(s, collections.abc.Iterable):
            site_list = list(s)
            if site_list and all(map(isa(PeriodicSite), site_list)):
                structure = from_pymatgen_structure(PymatgenStructure.from_sites(site_list))

    if isinstance(s, dict) and structure is None:
        if 'file' in s:
            structure = read_structure_from_file(settings)
        elif all(field in s for field in needed_fields):
            lattice = np.array(s['lattice'])
            coords = np.array(s['coords'])
            species = list(s['species'])
            structure = Structure(lattice, coords, species, (True, True, True))
        else:
            raise BadSettings(f'A structure dictionary needs the following fields {needed_fields}')

    if structure is None:
        raise BadSettings(f'Cannot read structure from the settings, "{type(s)}"')
    if isinstance(s, dict) and 'supercell' in s:
        sizes = settings.structure.supercell
        if len(sizes) != 3:
            raise BadSettings('To create a supercell you need to specify three lengths')
        structure = make_supercell(structure, *sizes)
        del settings.structure['supercell']

    return structure


@parameter('which', default=defaults.which, required=True)
def read_which(settings: AttrDict):
    structure = settings.structure
    if isinstance(settings.which, str):
        sublattice = settings.which
        allowed_sublattices = {'all', }.union(structure.unique_species)
        if sublattice not in allowed_sublattices:
            raise BadSettings(f'The structure does not have an "{sublattice}" sublattice. '
                              f'Possible values would be {allowed_sublattices}')
        if sublattice == 'all':
            mask = tuple(range(structure.num_atoms))
        else:
            mask = tuple(i for i, sp in enumerate(structure.species) if sp.symbol == sublattice)
        which = mask
    elif isinstance(settings.which, (list, tuple)):
        sublattice = tuple(set(settings.which))  # eliminate duplicates
        if len(sublattice) < 2:
            raise BadSettings('You need to at least specify two different lattice positions to define a sublattice')
        if not all(map(isa(int), sublattice)):
            raise BadSettings('I do only understand integer lists to specify a sublattice')
        if not all(map(lambda _: 0 <= _ < structure.num_atoms, sublattice)):
            raise BadSettings(f'All indices in the list must be 0 <= index < {structure.num_atoms}')
        which = tuple(settings.which)
    else:
        raise BadSettings('I do not understand your composition specification')
    # we remove the which clause from the settings, since we are replacing the structure with the sublattice structure
    return which


@parameter('composition', default=defaults.composition, required=True)
def read_composition(settings: AttrDict):
    structure = settings.structure[settings.which]
    if not isinstance(settings.composition, dict):
        raise BadSettings('Cannot interpret "composition" settings. I expect a dictionary')

    build_structure(settings.composition, structure)
    return settings.composition


@parameter('shell_distances', default=defaults.shell_distances, required=True)
def read_shell_distances(settings: AttrDict):
    if not isinstance(settings.shell_distances, (list, tuple, set)):
        raise BadSettings('I only understand list, sets, and tuples for the shell-distances parameter')
    if not all(map(isa(numbers.Real), settings.shell_distances)):
        raise BadSettings('All shell distances must be real values')
    distances = list(filter(lambda d: not np.isclose(d, 0.0), settings.shell_distances))
    if len(distances) < 1:
        raise BadSettings('You need to specify at least one shell-distance')
    for distance in distances:
        if distance < 0.0:
            raise BadSettings(f'A distance can never be less than zero. You specified "{distance}"')
    sorted_distances = list(sorted(distances))
    sorted_distances.insert(0, 0.0)
    return sorted_distances


@parameter('shell_weights', default=defaults.shell_weights, required=True)
def read_shell_weights(settings: AttrDict):
    if not isinstance(settings.shell_weights, dict):
        raise BadSettings('I only understand dictionaries for the shell-weight parameter')
    if len(settings.shell_weights) < 1:
        raise BadSettings('You have to include at least one coordination shell to carry out the optimization')
    allowed_indices = set(range(1, len(settings.shell_distances)))

    parsed_weights = {
        convert(shell, to=int, message=f'A shell must be an integer. You specified {shell}'):
            convert(weight, to=float, message=f'A weight must be a floating point number. You specified {weight}')
        for shell, weight in settings.shell_weights.items()
    }

    for shell in parsed_weights.keys():
        if shell not in allowed_indices:
            raise BadSettings(f'The shell {shell} you specified is not allowed. Allowed values are {allowed_indices}')
    return settings.shell_weights


def read_parameter_shaped_array(settings: AttrDict, parameter_name: str, constructor: T.Callable[[np.ndarray], np.ndarray]):
    nums = num_species(settings)
    nshells = num_shells(settings)

    if isinstance(settings.get(parameter_name), (list, tuple, np.ndarray)):
        w = np.array(settings.get(parameter_name)).astype(float)
        if w.ndim in {2, 3}:
            expected_shape = (nshells, nums, nums) if w.ndim == 3 else (nums, nums)

            ensure_array_shape(w, expected_shape, f'The 3D "{parameter_name}" you have specified '
                                                  f'has a wrong shape ({w.shape}). Expected {expected_shape}')
            ensure_array_symmetric(w, f'The "{parameter_name}" parameters are not symmetric')
            return constructor(w)

    raise BadSettings(f'As "{parameter_name}" I do expect a {nums}x{nums} matrix, '
                      f'since your structure contains {nums} different species')


@parameter('pair_weights', default=defaults.pair_weights, required=True)
def read_pair_weights(settings: AttrDict):
    sorted_weights = sorted(settings.shell_weights.items(), key=item(0))

    return read_parameter_shaped_array(settings, 'pair_weights',
                                       lambda w: w if w.ndim == 3
                                       else np.stack([w * shell_weight for _, shell_weight in sorted_weights]))


@parameter('target_objective', default=defaults.target_objective, required=True)
def read_target_objective(settings: AttrDict):
    if isinstance(settings.target_objective, (int, float)):
        target_objective = np.ones((num_shells(settings), num_species(settings), num_species(settings))) * settings.target_objective
        settings['target_objective'] = target_objective
    return read_parameter_shaped_array(settings, 'target_objective',
                                       lambda w: w if w.ndim == 3 else np.stack([w] * num_shells(settings)))


@parameter('prefactor_mode', default=defaults.prefactor_mode, required=True)
def read_prefactor_mode(settings: AttrDict):
    if not isinstance(settings.prefactor_mode, str):
        raise BadSettings('"prefactor_mode" must be of type str')
    if settings.prefactor_mode.lower() not in {'set', 'mul'}:
        raise BadSettings('"prefactor_mode" must be either {\'set\', \'mul\'}')
    return settings.prefactor_mode.lower()


@parameter('prefactors', default=defaults.prefactors, required=True)
def read_prefactors(settings: AttrDict):
    if isinstance(settings.prefactors, (int, float)):
        prefactors = np.ones((num_shells(settings), num_species(settings), num_species(settings))) * settings.prefactors
        settings['prefactors'] = prefactors
    prefactors = read_parameter_shaped_array(settings, 'prefactors',
                                       lambda w: w if w.ndim == 3 else np.stack([w] * num_shells(settings)))

    prefactors = prefactors if settings.prefactor_mode == 'set' else (prefactors * defaults.prefactors(settings))
    if np.any(np.isclose(prefactors, 0)):
        raise BadSettings('A prefactor of zero is not allowed as divison would lead to infinite numbers')
    return prefactors


@parameter('threads_per_rank', default=defaults.threads_per_rank, required=True)
def read_threads_per_rank(settings: AttrDict):
    converter = partial(convert, to=int, converter=int_safe, message='Cannot parse "threads_per_rank" argument')
    if isinstance(settings.threads_per_rank, (float, int)):
        return [converter(settings.threads_per_rank)]
    if isinstance(settings.threads_per_rank, (list, tuple, np.ndarray)):
        if len(settings.threads_per_rank) != 1:
            if not have_mpi_support():
                raise BadSettings('The module sqsgenerator.core.iteration was not compiled with MPI support')
        return list(map(converter, settings.threads_per_rank))

    raise BadSettings('Cannot interpret "threads_per_rank" setting.')


def process_settings(settings: AttrDict, params: T.Optional[T.Set[str]] = None, ignore: T.Iterable[str]=()) -> AttrDict:
    """
    Process an dict-like input parameters, according to the rules specified in the 
    `Input parameter documentation <https://sqsgenerator.readthedocs.io/en/latest/input_parameters.html>`_. This function
    should be used for processing user input. Therefore, exports the parser functions defined in 
    ``sqsgenerator.settings.readers``. To specify a specify subset of parameters the {params} argument is used.
    To {ignore} specific parameters pass a list of parameter names

    :param settings: the dict-like user configuration
    :type settings: AttrDict
    :param params: If specified only the subset of {params} is processed (default is ``None``)
    :type params: Optional[Set, None]
    :param ignore: a list/iterable of params to ignore (default is ``()``)
    :type ignore: Iterable[str]
    :return: the processed settings dictionary
    :rtype: AttrDict
    """
    
    params = params if params is not None else set(parameter_list())
    last_needed_parameter = max(params, key=parameter_index)
    ignore = set(ignore)
    for index, (param, processor) in enumerate(__parameter_registry.items()):
        if param not in params:
            # we can only skip this parameter if None of the other parameters depends on param
            if parameter_index(param) > parameter_index(last_needed_parameter):
                continue
        if param in ignore:
            continue
        settings[param] = processor(settings)
    return settings


def parameter_list() -> T.List[str]:
    return list(__parameter_registry.keys())


def parameter_index(p: str) -> int:
    return parameter_list().index(p)



"""
Analyse CLI command - Calculates SRO parameters, rank and objectives. Intended use is to analyze already existing
structures
"""

import os
import sys
import click
import functools
from rich import box
from rich.text import Text
from rich.table import Table
from sqsgenerator.io import dumps, to_dict
from sqsgenerator.core import rank_structure
from sqsgenerator.fallback.attrdict import AttrDict
from operator import attrgetter as attr, itemgetter as item
from sqsgenerator.commands.help import parameter_help as help, command_help
from sqsgenerator.settings import construct_settings, process_settings, defaults
from sqsgenerator.commands.common import click_settings_file, error, pretty_print
from sqsgenerator.public import pair_analysis, available_species, extract_structures, expand_sqs_results


species_to_ordinal = dict(map(attr('symbol', 'Z'), available_species()))


def foreach(f, iterable, unpack=False):
    for x in iterable:
        f(*(x if unpack else (x,)))


def map_values(f, d: dict, factory=dict):
    return factory({k: f(v) for k, v in d.items()})


def format_parameters(settings, result):
    """
    Nicely formats SQS parameters into a list of tables
    """
    parameters = result.parameters
    species = sorted(settings.structure.unique_species, key=species_to_ordinal.get)

    def make_table(shell, sw, sros, fmt='{0:.3f}'):
        table = Table(title=f'Shell {shell}, weight = {sw:.2f}', box=box.HORIZONTALS)
        table.add_column('Species', justify='center', style='bold')
        make_species_col = functools.partial(table.add_column, justify='center')
        foreach(make_species_col, species)
        for el, pars in zip(species, sros):
            table.add_row(el, *map(fmt.format, pars))
        return table

    sorted_shell_weights = sorted(settings.shell_weights.items(), key=item(0))

    return (make_table(shell, sw, sros) for (shell, sw), sros in zip(sorted_shell_weights, parameters))


@click.command('analyse', help=command_help.analyse)
@click.option('--output-format', '-of', type=click.Choice(['yaml', 'json', 'pickle', 'native']), default='native',
              help=help.output_format)
@click.option('--params', '-p', is_flag=True, help=help.dump_params)
@click_settings_file(process={'structure'}, ignore={'which'}, default_name='sqs.result.yaml')
def analyse(settings, output_format, params):
    if 'configurations' not in settings:
        # in this case we use the input from the loaded structure
        settings['configurations'] = {
            rank_structure(settings.structure): settings.structure.symbols.tolist()
        }
    # we do not want to override "which" upon parsing therefore we add it only once we it is not present
    if 'which' not in settings:
        settings['which'] = list(range(len(settings.structure)))
    num_configurations = len(settings.configurations)

    if num_configurations < 1:
        error('Your input does not contain any input configurations')

    # extract the defines sublattice from the merged structure
    structures = map_values(lambda s: s[settings.which], extract_structures(settings))

    # set one of the actual computes structures into settings, to get proper default values
    first_structure = next(iter(structures.values()))

    # a new clone of the actual settings object is create -> we do not want to loose the old -> TODO: find a reason why
    analyse_settings = AttrDict(settings.copy())
    analyse_settings.update(structure=first_structure)
    analyse_settings.update(which=defaults.which(analyse_settings))

    include_fields = ('objective', 'parameters', 'configuration')
    analyse_settings = process_settings(analyse_settings)
    # carry out the actual pair analysis by overwriting the structure in the current settings
    analyzed = map_values(lambda s: pair_analysis(construct_settings(analyse_settings, False, structure=s)), structures)
    document = expand_sqs_results(analyse_settings, list(analyzed.values()), fields=include_fields, inplace=params)

    if 'structure' in document:
        del document['structure']
    if 'which' in document:
        del document['which']
    if output_format == 'native':
        all_renderables = []
        for rank, result in map_values(AttrDict, document.configurations).items():
            renderables = list(format_parameters(analyse_settings, AttrDict(result)))
            renderables.insert(0, Text(f'Parameters:{os.linesep}', style='bold'))
            renderables.insert(0, Text.assemble(
                ('Configuration: ', 'bold'), (f'{result.configuration}', 'bold cyan'))
            )
            renderables.insert(0, Text.assemble(('Objective: ', 'bold'), (f'{result.objective}', 'bold cyan')))
            renderables.insert(0, Text.assemble(('Rank: ', 'bold'), (f'{rank}', 'bold cyan')))
            all_renderables.extend(renderables)
        pretty_print(*all_renderables)
    else:
        sys.stdout.buffer.write(dumps(to_dict(document), output_format=output_format))





class BadSettings(Exception):

    def __init__(self, message=None, parameter=None):
        super(BadSettings, self).__init__(message)
        self.message = message
        self.parameter = parameter



import click
from sqsgenerator.compat import available_features_with_version
from sqsgenerator.commands.run import run
from sqsgenerator.commands.export import export
from sqsgenerator.commands.params import params
from sqsgenerator.commands.analyse import analyse
from sqsgenerator.commands.compute import compute
from sqsgenerator.commands.common import pretty_format
from sqsgenerator.core import __version__, __features__


def make_version_string():
    major, minor, *_ = __version__
    return '%i.%i' % (major, minor)


def make_repo_status():
    _, _, commit, branch = __version__
    return '%s@%s' % (commit, branch)


_title = '[bold]sqsgenerator[/bold] - A CLI tool to find optimized SQS structures'


def make_version_message():
    return pretty_format(
        f"""{_title}
    [magenta italic]version:[/magenta italic] {make_version_string()}
    [magenta italic]build:[/magenta italic] {make_repo_status()}
    [magenta italic]features:[/magenta italic] {__features__}
    
    [magenta italic]author:[/magenta italic] Dominik Gehringer
    [magenta italic]email:[/magenta italic] dominik.gehringer@unileoben.ac.at
    [magenta italic]repo:[/magenta italic] https://github.com/dgehringer/sqsgenerator
    [magenta italic]docs:[/magenta italic] https://sqsgenerator.readthedocs.io/en/latest
    [magenta italic]modules:[/magenta italic] {available_features_with_version()}""")


@click.group(help=pretty_format(_title))
@click.version_option(prog_name='sqsgen', version=make_version_string(), message=make_version_message())
def cli():
    pass


cli.add_command(run, 'run')
cli.add_command(params, 'params')
cli.add_command(export, 'export')
cli.add_command(analyse, 'analyse')
cli.add_command(compute, 'compute')


if __name__ == '__main__':
    cli()



import logging
import operator
import functools
import frozendict
import typing as T
from sqsgenerator.core.core import SQSResult, Atom
from sqsgenerator.core.core import __version__, __features__
from sqsgenerator.core.core import set_log_level, pair_sqs_iteration, pair_analysis
from sqsgenerator.core.core import IterationMode, IterationSettings, BoostLogLevel
from sqsgenerator.core.structure import Structure, structure_to_dict, make_supercell
from sqsgenerator.core.core import default_shell_distances, atoms_from_numbers, atoms_from_symbols, available_species, \
    symbols_from_z, z_from_symbols, build_configuration, ALL_SITES, make_rank as make_rank_, \
    rank_structure as rank_structure_, total_permutations as total_permutations_, \
    compute_prefactors as compute_prefactors_

__all__ = [
    '__version__',
    '__features__',
    'set_log_level',
    'pair_sqs_iteration',
    'IterationMode',
    'IterationSettings',
    'BoostLogLevel',
    'pair_analysis',
    'Structure',
    'structure_to_dict',
    'get_function_logger',
    'Atom',
    'make_supercell',
    'SQSResult',
    'default_shell_distances',
    'total_permutations',
    'rank_structure',
    'atoms_from_numbers',
    'atoms_from_symbols',
    'available_species',
    'symbols_from_z',
    'z_from_symbols',
    'build_configuration',
    'ALL_SITES',
    'make_rank',
    'rank_structure',
]

attr = operator.attrgetter
item = operator.itemgetter
method = operator.methodcaller
partial = functools.partial

TRACE = 5
logging.addLevelName(TRACE, 'TRACE')

log_levels = frozendict.frozendict(
    trace=TRACE,
    debug=logging.DEBUG,
    info=logging.INFO,
    warning=logging.WARNING,
    error=logging.ERROR
)


def set_core_log_level(level: int = logging.WARNING) -> T.NoReturn:
    level_map = {
        TRACE: BoostLogLevel.trace,
        logging.DEBUG: BoostLogLevel.debug,
        logging.INFO: BoostLogLevel.info,
        logging.WARNING: BoostLogLevel.warning,
        logging.ERROR: BoostLogLevel.error
    }
    set_log_level(level_map.get(level))


def get_function_logger(f: T.Callable) -> logging.Logger:
    return logging.getLogger(f.__module__ + '.' + f.__name__)


set_core_log_level(logging.WARNING)


def total_permutations(struct: Structure, which: T.Optional[T.Iterable[int]] = None) -> int:
    """
    Compute the total number of total permutations, by looking at the configuration of {struct}. The number of
    of permutations are given by the multinomial coefficient. Let :math:`N_1, N_2, \ldots, N_m` the number of the
    atoms of the :math:`m^{\\text{th}}` species. Therefore, one can generate

    .. math::

        N_{\\text{perms}} = \dfrac{N!}{\prod_m^M N_m} \quad \\text{where} \quad \sum_m^M N_m = N

    permutations in total

    :param struct: the structure with the configuration to calculate the number of total permutations
    :type struct: Structure
    :param which: the indices of the lattice positions to choose (default is ``None``)
    :type which: Optional[Iterable[int]]
    :return: the rank of the structure of the sub-lattice selected by {which}
    :rtype: int
    """
    which = tuple(range(len(struct))) if which is None else which

    return total_permutations_(struct[which])


def make_rank(struct: Structure, configuration: T.Iterable[str], rank: int,
              which: T.Optional[T.Iterable[int]] = None) -> Structure:
    """
    Uses the lattice and positions from {struct} and generates the {rank}-th permutation (in lexicographical order)
    of {configuration} and and creates a new structure with it. If {which} is given we the selected lattices positions
    are used, therefore the length of {configuration} must match the length of {which}

    :param struct: the basis structure from which the new structure will be created
    :type struct: Structure
    :param configuration: the species sequence from which the {rank}-th permutation sequence
    :type configuration: Iterable[str]
    :param rank: the index of the permutation sequence
    :type rank: int
    :param which: the indices of the lattice positions to choose (default is ``None``)
    :type which: Optional[Iterable[int]]
    :raises IndexError: if the length of {which} does not match the length of {configuration}
    :raises ValueError: if the rank is less than 1 or greater than the number of :py:func:`total_permutations`
    :return: the structure with {rank}-th permutation sequence as configuration
    :rtype: Structure
    """

    msg = f"the rank of a configuration must me 1 <= rank <= total_permutations() = " \
          f"{total_permutations(struct.slice_with_species(configuration, which))}"
    if rank <= 0:
        raise ValueError(msg)

    which = tuple(range(len(struct))) if which is None else which
    try:
        unranked_configuration = make_rank_(configuration, rank)
    except IndexError:
        # the C++ extension raises an IndexError, we want to propagate it as an domain error
        raise ValueError(msg)

    return struct.with_species(unranked_configuration, which=which)


def rank_structure(struct: Structure, which: T.Optional[T.Iterable[int]] = None) -> int:
    """
    Computes the index (in lexicographical order) of the configuration sequence.
    If {which} is specified only those sites are selected to compute the permutation rank

    :param struct: the structure which carries the configuration sequence
    :type struct: Structure
    :param which: the indices of the lattice positions to choose (default is ``None``)
    :type which: Optional[Iterable[int]]
    :return: the index of the configuration sequence
    :rtype: int
    """
    which = tuple(range(len(struct))) if which is None else which
    return rank_structure_(struct[which])


def compute_prefactors(structure: Structure, composition: dict, shell_wights: dict, atol: float = 1e-3, rtol: float = 1e-5):
    return compute_prefactors_(structure, composition, shell_wights, atol, rtol)



import enum
import functools
import typing as T
from sqsgenerator.core import get_function_logger
from sqsgenerator.fallback.attrdict import AttrDict
from sqsgenerator.settings.exceptions import BadSettings


class Default(enum.Enum):
    """
    Dummy Enum to represent "no default" value. ``None`` is not possible, since it is a legit default value, we
    use  ``Default.NoDfault``to represent no default value
    """
    NoDefault = 0


def parameter(name: str, default: T.Optional[T.Any] = Default.NoDefault, required: T.Union[T.Callable, bool] = True,
              key: T.Union[T.Callable, str] = None, registry: T.Optional[dict] = None):
    if key is None: key = name
    get_required = lambda *_: required if isinstance(required, bool) else required
    get_key = lambda *_: key if isinstance(key, str) else key
    get_default = (lambda *_: default) if not callable(default) else default

    have_default = default != Default.NoDefault

    def _decorator(f: T.Callable):
        @functools.wraps(f)
        def _wrapped(settings: AttrDict):
            is_required = get_required(settings)
            k = get_key(settings)
            nonlocal name
            if k not in settings:
                if is_required:
                    if not have_default:
                        raise BadSettings(f'Required parameter "{name}" was not found', parameter=name)
                    else:
                        # a default is needed but found
                        df = get_default(settings)
                        get_function_logger(f).info(f'Parameter "{name}" was not found defaulting to: "{df}"')
                        return df
            else:
                # we catch the exception here and raise it again, to inject the parameter information automatically
                try: processed_value = f(settings)
                except BadSettings as exc:
                    exc.parameter = name # set the parameter info and forward the exception
                    raise
                else: return processed_value

        # register the parameters at the parameter_registry
        if registry is not None: registry[name] = _wrapped
        return _wrapped

    return _decorator


def if_(condition):
    def then_(th_val):
        def else_(el_val):
            def stmt_(settings):
                return th_val if condition(settings) else el_val
            return stmt_
        return else_
    return then_


def isa(o: T.Union[type]):
    """
    Constructs a type-checking predicate. Represents isinstance(x, o)
    :param o: the types to checked
    :type o: type
    :return: the predicate
    :rtype: Callable
    """
    return lambda x: isinstance(x, o)


def star(f: T.Callable) -> T.Callable:
    return lambda x: f(*x)


def const(o: T.Any) -> T.Callable:
    return lambda *_: o


def identity() -> T.Callable:
    return lambda _: _  # pragma: no cover




import numbers
import numpy as np
import typing as T
from functools import partial
from sqsgenerator.settings.exceptions import BadSettings
from sqsgenerator.core import symbols_from_z, z_from_symbols, Structure, ALL_SITES, build_configuration as build_configuration_


def ensure_array_shape(o: np.ndarray, shape: tuple, msg: T.Optional[str] = None):
    if o.shape != shape:
        raise (BadSettings(msg) if msg is not None else BadSettings)


def ensure_array_symmetric(o: np.ndarray, msg: T.Optional[str] = None):
    error = BadSettings(msg) if msg is not None else BadSettings
    if o.ndim == 2:
        if not np.allclose(o, o.T):
            raise error
    elif o.ndim == 3:
        for i, oo in enumerate(o, start=1):
            if not np.allclose(oo, oo.T):
                raise error
    return True


def int_safe(x):
    return int(float(x))


def convert(o, to=int, converter=None, on_fail=BadSettings, message=None):
    if isinstance(o, to):
        return o
    try:
        r = (to if converter is None else converter)(o)
    except (ValueError, TypeError):
        if on_fail is not None:
            raise on_fail(message.format(o)) if message is not None else on_fail()
    else:
        return r


def symbol_to_z(symbol: str):
    try:
        return next(iter(z_from_symbols([symbol])))
    except ValueError as e:
        # is raised by the extension, however we forward it as BadSettings Error
        raise BadSettings(e)


def to_internal_composition_specs(composition: dict, structure: Structure):
    initial_species = structure.unique_species
    parse_number_of_atoms = partial(convert, to=int, message='Failed to interpret "{0}" as a number of atoms')
    # there is the chance that a parser returns "0" species as integer, we have to consider this responsibilities

    def symbol_to_z_with_zero(k: T.Union[str, int]):
        if isinstance(k, int):
            if k == 0:
                return k
            else:
                raise BadSettings('I can only interpret "0" as an atomic species')
        else:
            return symbol_to_z(k)

    def parse_value(v: T.Union[int, dict]):
        if isinstance(v, numbers.Number):
            v = parse_number_of_atoms(v)
            if v < 1:
                raise BadSettings('You have distribute at least one atom')
            return {ALL_SITES: v}
        elif isinstance(v, dict):
            if not all(s in initial_species for s in v.keys()):
                raise BadSettings('You cannot distribute atoms on a sublattice which is not specified in the initial '
                                  'structure')
            return {symbol_to_z_with_zero(s): parse_number_of_atoms(n) for s, n in v.items()}
        else:
            raise BadSettings('Cannot interpret composition specification')

    return {symbol_to_z_with_zero(s): parse_value(v) for s, v in composition.items()}


def build_structure(composition: dict, structure: Structure):
    composition = to_internal_composition_specs(composition, structure)
    # after parsing we check if the internal routine runs without crashing, this is necessare since
    # "build_configuration" get's called in the IterationSettings constructor
    try:
        *_, conf = build_configuration_(structure.numbers.tolist(), composition)
    except (RuntimeError, ValueError) as e:
        # forward the error as BadSettings
        raise BadSettings(e)

    return structure.with_species(symbols_from_z(conf))



import typing as T
from sqsgenerator.core import IterationSettings
from sqsgenerator.fallback.attrdict import AttrDict
from sqsgenerator.settings.defaults import defaults
from sqsgenerator.settings.utils import build_structure
from sqsgenerator.settings.exceptions import BadSettings
from sqsgenerator.settings.readers import process_settings, parameter_list, to_internal_composition_specs


def construct_settings(settings: AttrDict, process: T.Optional[bool] = True, **overloads) -> IterationSettings:
    """
    Constructs a ``sqsgenerator.core.IterationSettings`` object. This object is needed to pass the {settings} to the
    C++ extension. This function does not modify the passed {settings} but creates a copy of it. This function is meant
    for **internal use**

    :param settings: the dict-like settings object. The parameter is passed on to the ``AttrDict`` constructor
    :type settings: AttrDict
    :param process: process the settings (default is ``True``)
    :type process: bool
    :param overloads: keyword args are used to **overload** key in the {settings} dictionary
    :return: the ``sqsgenerator.core.IterationSettings`` object, ready to pass on to ``pair_analysis``
    :rtype: IterationsSettings
    :raises KeyError: if a key was passed in {overloads} which is not present in {settings}
    """

    settings = AttrDict(settings.copy())
    for overload, value in overloads.items():
        if overload not in settings:
            raise KeyError(overload)
        settings[overload] = value
    settings = process_settings(settings) if process else settings

    try:
        return IterationSettings(
            settings.structure,
            to_internal_composition_specs(settings.composition, settings.structure),
            settings.target_objective,
            settings.pair_weights,
            settings.prefactors,
            dict(settings.shell_weights),
            settings.iterations,
            settings.max_output_configurations,
            list(settings.shell_distances),
            list(settings.threads_per_rank),
            settings.atol,
            settings.rtol,
            settings.mode
        )
    except (ValueError, RuntimeError) as e:
        raise BadSettings(e)


__all__ = [
    'BadSettings',
    'parameter_list',
    'process_settings',
    'construct_settings',
    'defaults',
    'build_structure'
]



"""
CLI command to perform a SQS iteration
"""

import os
import click
from sqsgenerator.commands.export import export_structures
from sqsgenerator.commands.common import click_settings_file
from sqsgenerator.io import to_dict, dumps, compression_to_file_extension
from sqsgenerator.public import log_levels, pair_sqs_iteration, expand_sqs_results
from sqsgenerator.commands.help import command_help as c_help, parameter_help as help


@click.command('iteration', help=c_help.iteration)
@click.option('--minimal/--no-minimal', '-m/-nm', default=True, help=help.minimal)
@click.option('--similar/--no-similar', '-s/-ns', default=False, help=help.similar)
@click.option('--log-level', type=click.Choice(list(log_levels.keys())), default='warning', help=help.log_level)
@click.option('--dump/--no-dump', '-d/-nd', default=True, help=help.dump)
@click.option('--export', '-e', 'do_export', is_flag=True, help=help.export)
@click.option('--dump-params', '-dp', is_flag=True, help=help.dump_params)
@click.option('--format', '-f', type=click.STRING, default='cif', help=help.format)
@click.option('--writer', '-w', type=click.Choice(['ase', 'pymatgen']), default='ase', help=help.writer)
@click.option('--output', '-o', 'output_file', type=click.Path(dir_okay=False, allow_dash=True), help=help.output_file)
@click.option('--dump-format', '-df', 'dump_format', type=click.Choice(['yaml', 'json', 'pickle']), default='yaml',
              help=help.dump_format)
@click.option('--dump-include', '-di', 'dump_include', type=click.Choice(['parameters', 'timings', 'objective']),
              multiple=True, help=help.dump_include)
@click.option('--compress', '-c', type=click.Choice(list(compression_to_file_extension)), help=help.compress)
@click_settings_file('all')
def iteration(settings, minimal, similar, log_level, dump, dump_params, do_export, format, writer, output_file,
              dump_format, dump_include, compress):
    sqs_results, timings = pair_sqs_iteration(settings, minimal=minimal, similar=similar, log_level=log_level)

    final_document = expand_sqs_results(settings, sqs_results, timings, fields=dump_include, inplace=dump_params)
    output_prefix = output_file \
        if output_file is not None \
        else (os.path.splitext(settings.file_name)[0] if 'file_name' in settings else 'sqs')

    if dump:
        output_file_name = f'{output_prefix}.result.{dump_format}'
        with open(output_file_name, 'wb') as handle:
            handle.write(
                dumps(
                    to_dict(final_document), output_format=dump_format
                ))

    if do_export:
        export_structures(settings, final_document, format=format, writer=writer, output_file=output_prefix,
                          compress=compress)
    return final_document


@click.group('run', help=c_help.iteration)
def run():
    pass


run.add_command(iteration, 'iteration')



"""
Command to check the actual parameters which are going to be used in a SQS iteration
"""

import sys
import click
from sqsgenerator.io import dumps, to_dict
from sqsgenerator.commands.common import pretty_print, click_settings_file
from sqsgenerator.commands.help import command_help as c_help, parameter_help as help
from sqsgenerator.settings import process_settings, parameter_list, construct_settings, build_structure


@click.command('show', help=c_help.params.show)
@click.option('--param', '-p', type=click.Choice(parameter_list()), default=None, multiple=True, help=help.param)
@click.option('--output-format', '-of', type=click.Choice(['yaml', 'json', 'pickle', 'native']), default='native',
              help=help.output_format)
@click.option('--inplace', is_flag=True, help=help.inplace)
@click_settings_file(process=None)
def show(settings, param=None, output_format='yaml', inplace=False):
    parameters_selected = bool(param)
    parameters_to_process = set(param or parameter_list())
    processed = process_settings(settings, params=parameters_to_process)
    parameters_to_output = parameters_to_process.union({} if parameters_selected else {'sublattice'})
    filtered = {k: v for k, v in to_dict(processed).items() if k in parameters_to_output}

    if inplace:
        output_format = settings.input_format
        output_file = open(settings.file_name, 'wb')
    else:
        output_file = sys.stdout.buffer

    if output_format == 'native':
        pretty_print(filtered)
    else:
        output_file.write(dumps(filtered, output_format=output_format))

    if inplace:
        output_file.close()


@click.command('check', help=c_help.params.check)
@click_settings_file('all')
def check(settings):
    construct_settings(
        settings,
        process=False,
        structure=build_structure(settings.composition, settings.structure[settings.which])
    )


@click.group('params', help=c_help.params.command)
def params(*args, **kwargs):
    pass


params.add_command(show, 'show')
params.add_command(check, 'check')




import sqsgenerator.cli
import sqsgenerator.public

from sqsgenerator.public import process_settings, sqs_optimize, from_pymatgen_structure, from_ase_atoms, \
    to_ase_atoms, to_pymatgen_structure, IterationMode, Structure, read_settings_file, export_structures, make_rank, \
    rank_structure, sqs_analyse, total_permutations

__all__ = [
    'process_settings',
    'sqs_optimize',
    'from_ase_atoms',
    'from_pymatgen_structure',
    'to_ase_atoms',
    'to_pymatgen_structure',
    'IterationMode',
    'Structure',
    'read_settings_file',
    'export_structures',
    'make_rank',
    'rank_structure',
    'sqs_analyse',
    'total_permutations'
]